// main.cpp
// Copyright (C) 2011 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Heat.h"

#include <Basic/Log.h>
#include <Basic/Macros.h>
#include <Container/CSRMatrix.h>
#include <Container/Vector.h>
#include <Communication/MPI.h>
#include <DomainDecomposition/SchurComplement.h>
#include <FiniteElement/Assembler.h>
#include <FiniteElement/Geometry.h>
#include <FiniteElement/ShapeFunctions.h>
#include <Solver/SolverParameters.h>
#include <Partition/Partition.h>

#include <omp.h>


enum
{
	tag_dyn_steps,
	tag_dynamic_C_count,
	tag_dynamic_C_index,
	tag_dynamic_C_entry,
	tag_dynamic_ui,
	tag_dynamic_Cui
};


#define VERSION "1.0.0"
#define PREFIX_MAXIMUM_LENGTH 900
#define PATH_MAXIMUM_LENGTH 1000


int main(int argc, char** argv)
{
	if ((argc < 2) || (argc > 4))
	{
		fprintf(stderr, "Invalid number of arguments. Use:\n  %s <problem_files_prefix> [log_level] [log_file]\n", argv[0]);
		fprintf(stderr, " Example:\n  %s gid_examples/problem.gid/problem\n\n", argv[0]);
		fputs(" Required problem files are:\n", stderr);
		fputs("  <problem_file_prefix>.solver.dat\n", stderr);
		fputs("  <problem_file_prefix>.geometry.dat\n", stderr);
		fputs("  <problem_file_prefix>.problem.dat\n\n", stderr);
		fputs(" By default, log level is set to 2:\n", stderr);
		return 1;
	}

	char solver_file_name[PATH_MAXIMUM_LENGTH];
	char geometry_file_name[PATH_MAXIMUM_LENGTH];
	char problem_file_name[PATH_MAXIMUM_LENGTH];
	char results_file_name[PATH_MAXIMUM_LENGTH];

	sprintf(solver_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.solver.dat", argv[1]);
	sprintf(geometry_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.geometry.dat", argv[1]);
	sprintf(problem_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.problem.dat", argv[1]);

	log_level = (argc > 2) ? atoi(argv[2]) : 2;
	if (argc == 4)
	{
		log_stream = fopen(argv[3], "wb");
		if (!log_stream)
		{
			log_stream = stdout;
			fprintf(stderr, "[Warning] Failed to create log file: %s", argv[3]);
		}
	}

	try
	{
		// Load solver parameters
		SolverParameters<double> solver_parameters(solver_file_name);

		// MPI object
		MPI mpi(argc, argv);
		if (mpi.size < 3)
		{
			if (mpi.rank == 0)
			{
				fputs("[Error] Invalid number processes to start, minimum 3 processes should be started.\n", stderr);
			}
			return 1;
		}

		if (mpi.rank == 0) // Master
		{
			int partitions_count = mpi.size - 1;

			Log(1, "Heat.MPI ------------------------------------------------------------");
			Log(1, "-Version:       " VERSION);
			Log(1, "-Solver file:   %s", solver_file_name);
			Log(1, "-Geometry file: %s", geometry_file_name);
			Log(1, "-Problem file:  %s", problem_file_name);
			Log(1, "-Partitions:    %i", partitions_count);

			// Print solver parameters
			solver_parameters.PrintInfo();

			// Load geometry
			Geometry<double> geometry(geometry_file_name);
			geometry.PrintInfo();

			// Partition mesh
			Vector<Substructure> substructures;
			Boundary boundary;
			StructurePartitioning(geometry.mesh, partitions_count, solver_parameters.reorder, substructures, boundary);
			Log(1, "Partitioning done");

			// Set shape functions
			ShapeFunctions<double> shape_functions(geometry.mesh, geometry.nodes);

			// Load problem parameters
			Heat<double> heat(problem_file_name, shape_functions, geometry.material_index);

			// Boundary geometry
			const Vector<int>& element_index = boundary.element_index;
			const Vector<int>& node_index = boundary.node_index;

			SchurComplementMaster<double> schur_complement_master(mpi, substructures, boundary, 1);

			// Assembler class for boundaries
			Assembler assemblerB(geometry.mesh, element_index, node_index, 1);

if (heat.problem_type == time_independent)
{
			// Assemble and send systems of equations and links
			Log(1, "Systems of equations -------------------------------------------------");

			// Assemble global vectors (temperatures, fluxes and fixed)
			int M = geometry.mesh.nodes_count;
			Vector<double> global_u(M);
			Vector<double> global_f(M);
			Vector<bool> global_fixed(M);

			// Fill global vectors
			heat.FillU(global_u);
			heat.FillFixed(global_fixed);
			heat.FillF(global_f);

			// Elemental matriz size
			int N = geometry.mesh.nodes_per_element;

			// Allocate KBB, uB, fB and fixedB
			int B = boundary.node_index.size;
			Matrix<double> KBB(B, B);
			Vector<double> uB(B);
			Vector<double> fB(B);
			Vector<bool> fixedB(B);

			// Assemble KBB
			KBB.Fill(0.0);
			Matrix<double> Ke(N, N);
			for (int i = 1; i <= element_index.size; ++i)
			{
				int element_id = element_index.entry[i];

				heat.FillKe(element_id, Ke);
				assemblerB.AssembleAe(element_id, Ke, KBB);
			}

			// Assemble uB
			assemblerB.AssembleV(global_u, uB);

			// Assemble fB
			assemblerB.AssembleV(global_f, fB);

			// Assemble fixedB
			assemblerB.AssembleV(global_fixed, fixedB);

			// Systems of equations
			#pragma omp parallel for default(shared) schedule(dynamic,1) num_threads(solver_parameters.decomposition_threads)
			for (int p = 1; p <= partitions_count; ++p)
			{
				int steps = 0;
				mpi.Send(p, tag_dyn_steps, steps);

				const Vector<int>& element_index = substructures.entry[p].element_index;
				const Vector<int>& node_index = substructures.entry[p].node_index;

				// Assembler class
				Assembler assembler(geometry.mesh, element_index, node_index, 1);

				// Matrices size
				int Mi = node_index.size; // Matrix size
				int Bi = substructures.entry[p].boundary_count;
				int Ii = Mi - Bi;

				// Assemble K
				CSRMatrix<double> K(Mi, Mi);
				assembler.AllocateMatrix(K);
				K.Fill(0);
				Matrix<double> Ke(N, N);
				for (int i = 1; i <= element_index.size; ++i)
				{
					int element_id = element_index.entry[i];

					heat.FillKe(element_id, Ke);
					assembler.AssembleAe(element_id, Ke, K);
				}

				// Assemble u
				Vector<double> u(Mi);
				assembler.AssembleV(global_u, u);

				// Assemble f
				Vector<double> f(Mi);
				assembler.AssembleV(global_f, f);

				// Assemble fixed
				Vector<bool> fixed(Mi);
				assembler.AssembleV(global_fixed, fixed);

				// To avoid double compensation on fixed conditions on boundaries
				for (int i = Ii + 1; i <= Mi; ++i)
				{
					fixed.entry[i] = false;
				}

				// Apply fixed conditions
				Solver<double> fixer(K, u, f, fixed, false, 1);
				fixer.CompensateFixed();

				// Initialize Schur method for this partition
				schur_complement_master.SendSystemOfEquations(p, K, u, f);

				Log(1, "Partition %i:", p);
				Log(1, "-Degrees of freedom: %i", Mi);
				Log(1, "-nnz(K):             %i", K.NonZero());
			}

			// Schur complement method
			Log(1, "Schur complement method ----------------------------------------------");
			Log(1, "-Degrees of freedom: %i", B);
			Log(1, "-Threads:            %i", solver_parameters.decomposition_threads);
			Log(1, "-Multiple results:   %s", solver_parameters.multiple_results ? "yes" : "no");

			schur_complement_master.CalculateSchurComplement(KBB, uB, fB, fixedB, solver_parameters.decomposition_threads);
			schur_complement_master.CompleteSolution(uB, global_u);

			// Save results
			Log(1, "Save results ---------------------------------------------------------");
			if (solver_parameters.multiple_results)
			{
				char mesh_file_name[PATH_MAXIMUM_LENGTH];
				char mesh_name[31];

				for (int p = 1; p <= partitions_count; ++p)
				{
					Vector<int>& node_index = substructures.entry[p].node_index;
					Vector<int>& element_index = substructures.entry[p].element_index;

					if (heat.save_mesh)
					{
						sprintf(mesh_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.%03i.post.msh", argv[1], p);
						sprintf(mesh_name, "Solid_%03i", p);
						geometry.SaveMesh(mesh_file_name, mesh_name, element_index, node_index, p);
						Log(1, "Mesh saved: %s", mesh_file_name);
					}

					Vector<double> u(node_index.size);
					for (int l = 1; l <= node_index.size; ++l)
					{
						int n = node_index.entry[l];
						u.entry[l] = global_u.entry[n];
					}

					sprintf(results_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.%03i.post.res", argv[1], p);
					File results_file;
					results_file.Create(results_file_name);
					results_file.Write("GiD Post Result File 1.0\n\n");
					heat.WriteFluxGaussPoints(results_file);
					if (heat.calculate_temperature)
					{
						heat.WriteTemperatures(results_file, node_index, u, 1);
					}
					if (heat.calculate_flux)
					{
						heat.WriteFlux(results_file, element_index, node_index, u, 1);
					}
					results_file.Close();
					Log(1, "Results saved: %s", results_file_name);
				}
			}
			else // Save single file results
			{
				Vector<int> global_element_index(geometry.mesh.elements_count);
				global_element_index.FillSeries(1, 1);

				Vector<int> global_node_index(geometry.mesh.nodes_count);
				global_node_index.FillSeries(1, 1);

				// Save mesh
				if (heat.save_mesh)
				{
					char mesh_file_name[PATH_MAXIMUM_LENGTH];

					sprintf(mesh_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.post.msh", argv[1]);
					geometry.SaveMesh(mesh_file_name, "Heat", global_element_index, global_node_index, 1);
					Log(1, "Mesh saved: %s", mesh_file_name);
				}

				// Save results
				sprintf(results_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.post.res", argv[1]);
				File results_file;
				results_file.Create(results_file_name);
				results_file.Write("GiD Post Result File 1.0\n\n");
				heat.WriteFluxGaussPoints(results_file);
				if (heat.calculate_temperature)
				{
					heat.WriteTemperatures(results_file, global_node_index, global_u, 1);
				}
				if (heat.calculate_flux)
				{
					heat.WriteFlux(results_file, global_element_index, global_node_index, global_u, 1);
				}
				results_file.Close();
				Log(1, "Results saved: %s", results_file_name);
			}
}







else // time_dependent
{
	// R. W. Lewis, P. Nithiarasu, K. N. Seetharamu
	// Fundamentals of the Finite Element Method for Heat and Fluid Flow
	// Wiley, 2004, pp 156-160

	Log(1, "Systems of equations -------------------------------------------------");

	// Matrix size
	int M = geometry.mesh.nodes_count;

	// Elemental matriz size
	int N = geometry.mesh.nodes_per_element;

	// Allocate and fill global_u
	Vector<double> global_u(M);
	heat.FillU(global_u);

	// Allocate and fill global_f
	Vector<double> global_f(M);
	heat.FillF(global_f);

	// Allocate and fill global_fixed
	Vector<bool> global_fixed(M);
	heat.FillFixed(global_fixed);

	// Allocate ABB
	int B = boundary.node_index.size;
	Matrix<double> ABB(B, B);
	ABB.Fill(0.0);

	// Allocate CBB
	CSRMatrix<double> CBB(B, B);
	assemblerB.AllocateMatrix(CBB);
	CBB.Fill(0.0);

	// Assemble ABB and CBB
	Matrix<double> Ke(N, N);
	Matrix<double> Ce(N, N);
	for (int i = 1; i <= element_index.size; ++i)
	{
		int element_id = element_index.entry[i];

		heat.FillKeCe(element_id, Ke, Ce);
		assemblerB.AssembleAe(element_id, Ke, ABB);
		assemblerB.AssembleAe(element_id, Ce, CBB);
	}
	for (int i = 1; i <= M; ++i)
	{
		int k_max = CBB.count[i];
		for (int k = 1; k <= k_max; ++k)
		{
			int j = CBB.index[i][k];
			double a = ABB.entry[i][j];
			double c = CBB.entry[i][k];
			ABB.entry[i][j] = c + heat.time_scheme_factor*heat.time_per_step*a;
			CBB.entry[i][k] = c - (1 - heat.time_scheme_factor)*heat.time_per_step*a;
		}
	}

	// Allocate and assemble uB
	Vector<double> uB(B);
	assemblerB.AssembleV(global_u, uB);

	// Allocate and assemble fB
	Vector<double> fB(B);
	assemblerB.AssembleV(global_f, fB);

	// Allocate and assemble fixedB
	Vector<bool> fixedB(B);
	assemblerB.AssembleV(global_fixed, fixedB);

	Vector<Assembler*> partition_assembler(partitions_count);
	partition_assembler.Fill((Assembler*)0);

	// Systems of equations
	#pragma omp parallel for default(shared) schedule(dynamic,1) num_threads(solver_parameters.decomposition_threads)
	for (int p = 1; p <= partitions_count; ++p)
	{
		mpi.Send(p, tag_dyn_steps, heat.steps);

		const Vector<int>& element_index = substructures.entry[p].element_index;
		const Vector<int>& node_index = substructures.entry[p].node_index;

		// Assembler class
		partition_assembler.entry[p] = new Assembler(geometry.mesh, element_index, node_index, 1);
		if (!partition_assembler.entry[p])
		{
			throw Memory::Exception();
		}
		Assembler& assembler = *partition_assembler.entry[p];

		// Matrices size
		int Mi = node_index.size; // Matrix size
		int Bi = substructures.entry[p].boundary_count;
		int Ii = Mi - Bi;

		// Allocate A
		CSRMatrix<double> A(Mi, Mi);
		assembler.AllocateMatrix(A);
		A.Fill(0);

		// Allocate C
		CSRMatrix<double> C(Mi, Mi);
		C.CopyStructure(A);
		C.Fill(0);

		// Assemble A and C
		Matrix<double> Ke(N, N);
		Matrix<double> Ce(N, N);
		for (int i = 1; i <= element_index.size; ++i)
		{
			int element_id = element_index.entry[i];

			heat.FillKeCe(element_id, Ke, Ce);
			assembler.AssembleAe(element_id, Ke, ABB);
			assembler.AssembleAe(element_id, Ce, CBB);
		}
		for (int i = 1; i <= M; ++i)
		{
			int k_max = A.count[i];
			for (int k = 1; k <= k_max; ++k)
			{
				double a = A.entry[i][k];
				double c = C.entry[i][k];
				A.entry[i][k] = c + heat.time_scheme_factor*heat.time_per_step*a;
				C.entry[i][k] = c - (1 - heat.time_scheme_factor)*heat.time_per_step*a;
			}
		}

		// Assemble u
		Vector<double> u(Mi);
		assembler.AssembleV(global_u, u);

		// Assemble f
		Vector<double> f(Mi);
		assembler.AssembleV(global_f, f);

		// Assemble fixed
		Vector<bool> fixed(Mi);
		assembler.AssembleV(global_fixed, fixed);

		// To avoid double compensation on fixed conditions on boundaries
		for (int i = Ii + 1; i <= Mi; ++i)
		{
			fixed.entry[i] = false;
		}

		// Apply fixed conditions
		Solver<double> fixer(A, u, f, fixed, false, 1);
		fixer.CompensateFixed();

		// Initialize Schur method for this partition
		schur_complement_master.SendSystemOfEquations(p, A, u, f);

		// Send C matrix
		schur_complement_master.SendMatrix(p, tag_dynamic_C_count, tag_dynamic_C_index, tag_dynamic_C_entry, C);
		
		Log(1, "Partition %i:", p);
		Log(1, "-Degrees of freedom: %i", Mi);
		Log(1, "-nnz(A):             %i", A.NonZero());
		Log(1, "-nnz(C):             %i", C.NonZero());
	}

	// Schur complement method
	Log(1, "Schur complement method ----------------------------------------------");
	Log(1, "-Degrees of freedom: %i", B);
	Log(1, "-Threads:            %i", solver_parameters.decomposition_threads);
	Log(1, "-Multiple results:   %s", solver_parameters.multiple_results ? "yes" : "no");

	schur_complement_master.CalculateSchurComplement(ABB, uB, fB, fixedB, solver_parameters.decomposition_threads);
	schur_complement_master.CompleteSolution(uB, global_u);

	Log(1, "Time steps -----------------------------------------------------------");

	heat.time = 0;
	Vector<double> global_f_previous(global_f); // Previous f
	Vector<double> b(M);

	// Iterate and save results
	File results_file;
	results_file.Create(results_file_name);
	results_file.Write("GiD Post Result File 1.0\n\n");
	heat.WriteFluxGaussPoints(results_file);


// <- HERE

	for (int step = 1; step <= heat.steps; ++step)
	{
		// Set time
		heat.time = heat.time_per_step*step;

		// Fill global_f
		heat.FillF(global_f);

		// Calculate b
		b.Fill(0);
		for (int i = 1; i <= boundary.node_index.size; ++i)
		{
		}

		for (int p = 1; p <= partitions_count; ++p)
		{
			int Mi = substructures.entry[p].node_index.size;
			Vector<double> ui(Mi);
			partition_assembler.entry[p]->AssembleV(global_u, ui);
			mpi.Send(p, tag_dynamic_ui, ui);
			mpi.Receive(p, tag_dynamic_Cui, ui);
			for (int i = 1; i <= Mi; ++i)
			{
				int n = ubstructures.entry[p].node_index.entry[i];
				b.entry[n] += ui.entry[i];
			}
		}
/*
		// Calculate b
		for (int i = 1; i <= M; ++i)
		{
			double sum = 0;
			int k_max = C.count[i];
			for (int k = 1; k <= k_max; ++k)
			{
				sum += C.entry[i][k]*u.entry[C.index[i][k]];
			}
			b.entry[i] = sum + heat.time_per_step*(heat.time_scheme_factor*global_f.entry[i] + (1 - heat.time_scheme_factor)*f_previous.entry[i]);
			f_previous.entry[i] = global_f.entry[i];
		}

		// Assemble u
		heat.FillU(global_u);
		assembler.AssembleV(global_u, u);
		solver->CompensateFixed();

		// Solve system of equations
		if (!solver->Calculate())
		{
			Log(1, "Solution: INVALID");
			break;
		}
*/
		if ((step - 1) % heat.result_every_steps == 0)
		{
			if (heat.calculate_temperature)
			{
				heat.WriteTemperatures(results_file, node_index, global_u, step);
			}
			if (heat.calculate_flux)
			{
				heat.WriteFlux(results_file, element_index, node_index, global_u, step);
			}
		}
		Log(2, "Step %i", step);
	}
	results_file.Close();
	Log(1, "Results saved: %s", results_file_name);

	// Save mesh
	if (heat.save_mesh)
	{
		char mesh_file_name[PATH_MAXIMUM_LENGTH];

		sprintf(mesh_file_name, "%." MacroValueToString(PREFIX_MAXIMUM_LENGTH) "s.post.msh", argv[1]);
		geometry.SaveMesh(mesh_file_name, "Solid", element_index, node_index, 1);
		Log(1, "Mesh saved: %s", mesh_file_name);
	}

	// Delete assemblers data
	for (int p = 1; p <= partitions_count; ++p)
	{
		delete partition_assembler.entry[p];
	}


}

			Log(1, "Peak allocated memory ------------------------------------------------");
			Log(1, "-Master:        %12lu bytes", (unsigned long)Memory::peak_usage);
			Vector<size_t> partition_memory_peak_usage(partitions_count);
			schur_complement_master.GetMemoryPeakUsage(partition_memory_peak_usage);
			size_t total_memory_peak_usage = Memory::peak_usage;
			for (int p = 1; p <= partitions_count; ++p)
			{
				total_memory_peak_usage += partition_memory_peak_usage.entry[p];
				Log(1, "-Partition %3i: %12lu bytes", p, (unsigned long)partition_memory_peak_usage.entry[p]);
			}
			Log(1, "-Total:         %12lu bytes", (unsigned long)total_memory_peak_usage);
		}
		else // rank != 0
		{
			log_level = 0; // Prevents slaves logs to appear

			int steps;
			mpi.Receive(0, tag_dyn_steps, steps);

			SchurComplementSlave<double> schur_complement_slave(mpi, solver_parameters);
			schur_complement_slave.Initialize();

			if (steps == 0)
			{
				schur_complement_slave.CompleteSolution();
			}
			else
			{
				CSRMatrix<double> CIIi;
				CSRMatrix<double> CIBi;
				CSRMatrix<double> CBIi;
				schur_complement_slave.ReceiveMatrix(tag_dynamic_C_count, tag_dynamic_C_index, tag_dynamic_C_entry, CIIi, CIBi, CBIi);

				Vector<double> uIi(CIIi.columns);
				Vector<double> uBi(CIBi.columns);
				Vector<double> vIi(CIIi.rows);
				Vector<double> vBi(CIBi.rows);
				for (int step = 1; step <= steps; ++step)
				{
					schur_complement_slave.ReceiveVector(tag_dynamic_ui, uIi, uBi);
					for (int i = 1; i <= CIIi.rows; ++i)
					{
						double sum = 0;
						int k_max = CIIi.count[i];
						for (int k = 1; k <= k_max; ++k)
						{
							int j = CIIi.index[i][k];
							sum += CIIi.entry[i][k]*uIi.entry[j];
						}
						k_max = CIBi.count[i];
						for (int k = 1; k <= k_max; ++k)
						{
							int j = CIBi.index[i][k];
							sum += CIBi.entry[i][k]*uBi.entry[j];
						}
						vIi.entry[i] = sum;
					}
					for (int i = 1; i <= CBIi.rows; ++i)
					{
						double sum = 0;
						int k_max = CBIi.count[i];
						for (int k = 1; k <= k_max; ++k)
						{
							int j = CBIi.index[i][k];
							sum += CBIi.entry[i][k]*uIi.entry[j];
						}
						vBi.entry[i] = sum;
					}
					schur_complement_slave.SendVector(tag_dynamic_Cui, vIi, vBi);

					schur_complement_slave.CompleteSolution();
				}
			}

			schur_complement_slave.ReportMemoryPeakUsage();
		}
	}
	catch (Memory::Exception& exception)
	{
		fprintf(stderr, "[Error] Not enough memory: %s", exception.message);
	}
	catch (File::ExceptionCreate& exception)
	{
		fprintf(stderr, "[Error] Creating file: %s", exception.message);
	}
	catch (File::ExceptionEOF& exception)
	{
		fprintf(stderr, "[Error] Unexpected EOF: %s", exception.message);
	}
	catch (File::ExceptionFormat& exception)
	{
		fprintf(stderr, "[Error] Incorrect format file: %s", exception.message);
	}
	catch (File::ExceptionOpen& exception)
	{
		fprintf(stderr, "[Error] Opening file: %s", exception.message);
	}
	catch (File::ExceptionRead& exception)
	{
		fprintf(stderr, "[Error] Reading file: %s", exception.message);
	}
	catch (File::ExceptionSeek& exception)
	{
		fprintf(stderr, "[Error] Seeking in file: %s", exception.message);
	}
	catch (File::ExceptionWrite& exception)
	{
		fprintf(stderr, "[Error] Writing file: %s", exception.message);
	}
	catch (MPI::Exception& exception)
	{
		fprintf(stderr, "[Error] MPI exception: %s", exception.message);
	}

	if (Memory::current_usage != 0)
	{
		fprintf(stderr, "[Error] Memory leak: %lu bytes", Memory::current_usage);
	}

	if (log_stream != stdout)
	{
		fclose(log_stream);
	}

	return 0;
}
